/*
 *  Fast SD read with no CRC check
 *  Copyright (C) 2018 Sandor Zsuga (Jubatian)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef FASTREAD_H
#define FASTREAD_H


#include <bootlib.h>


/*
** Sends SD command and waits for the first response byte (normally R1), which
** it returns. Supports low SPI speeds. Pulls CS low before sending the
** command, and keeps it that way (low).
*/
uint8_t  SDC_Command_Nc(uint8_t cmd, uint32_t data);

/*
** Performs a single sector read (attempts a retry on fault), no CRC check
** on data.
**
** You can use it in combination with the filesystem, replacing an
** FS_Read_Sector() call as follows:
**
** SDC_Read_Sector_Nc(&sd_struct, FS_Get_Sector(&sd_struct));
**
** Returns zero on success, otherwise:
** 2: CMD17 failed
** 3: Timed out during waiting for data token
*/
uint8_t  SDC_Read_Sector_Nc(sdc_struct_t* sds, uint32_t sector);


/*
** Performs a single sector read without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** r23:r22: 512b sector address, high
** r21:r20: 512b sector address, low (together they are a proper C uint32)
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          2: CMD17 failed
**          3: Timed out during waiting for data token
** Clobbers:
** r0, r20, r21, r22, r23, r24, r25, Z, T(SREG)
*/
uint8_t  SD_Read_Sector_Nr_Nc(sdc_struct_t* sds, uint32_t sector);


/*
** Cues a single sector read without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** r23:r22: 512b sector address, high
** r21:r20: 512b sector address, low (together they are a proper C uint32)
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          2: CMD17 failed
** Clobbers:
** r0, r20, r21, r22, r23, r24, r25, Z, T(SREG)
*/
uint8_t  SD_Cue_Sector_Nr_Nc(sdc_struct_t* sds, uint32_t sector);


/*
** Completes a cued single sector read without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          3: Timed out during waiting for data token
** Clobbers:
** r0, r20, r21, r22, r23, r24, r25, Z, T(SREG)
*/
uint8_t  SD_CompleteCued_Sector_Nr_Nc(sdc_struct_t* sds);


#endif
