/*
 *  Fast SD read with no CRC check
 *  Copyright (C) 2018 Sandor Zsuga (Jubatian)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <avr/io.h>


.section .text



/*
** Ports and pins
*/
#define SPI_P   _SFR_IO_ADDR(PORTB)
#define SPI_DDR _SFR_IO_ADDR(DDRB)
#define CS_P    _SFR_IO_ADDR(PORTD)
#define CS_DDR  _SFR_IO_ADDR(DDRD)
#define SD_SCK  7
#define SD_MOSI 5
#define SD_MISO 6
#define SD_CS   6

/*
** SPI registers
*/
#define SPI_CR  _SFR_IO_ADDR(SPCR)
#define SPI_DR  _SFR_IO_ADDR(SPDR)
#define SPI_SR  _SFR_IO_ADDR(SPSR)



/*
** Sends SD command and waits for the first response byte (normally R1), which
** it returns. Supports low SPI speeds. Pulls CS low before sending the
** command, and keeps it that way (low).
**
** Inputs:
**     r24: Command byte
** r23:r22: Argument, high
** r21:r20: Argument, low (together they are a proper C uint32)
** Outputs:
**     r24: First response byte (normally R1)
** Clobbers:
** r0, r24, r25
*/
.global SDC_Command_Nc
SDC_Command_Nc:

	cbi   CS_P,    SD_CS   ; Chip Select: Low
	andi  r24,     0x3F
	ori   r24,     0x40    ; Form command byte
	rcall sdlib_wait_spi_with_FF
	out   SPI_DR,  r24     ; Send command byte
	rcall sdlib_wait_spi
	out   SPI_DR,  r23     ; Send arg. byte 0
	rcall sdlib_wait_spi
	out   SPI_DR,  r22     ; Send arg. byte 1
	rcall sdlib_wait_spi
	out   SPI_DR,  r21     ; Send arg. byte 2
	rcall sdlib_wait_spi
	out   SPI_DR,  r20     ; Send arg. byte 3
	rcall sdlib_wait_spi
	out   SPI_DR,  r24     ; Send CRC
	rcall sdlib_wait_spi
	ldi   r25,     16      ; Wait up to 16 bytes for response
SD_Command_wl:
	rcall sdlib_wait_spi_with_FF
	in    r24,     SPI_DR
	cpi   r24,     0xFF
	brne  SD_Command_ret
	dec   r25
	brne  SD_Command_wl
SD_Command_ret:
	ret


/*
** Cues a single sector read without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** r23:r22: 512b sector address, high
** r21:r20: 512b sector address, low (together they are a proper C uint32)
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          2: CMD17 failed
** Clobbers:
** r0, r20, r21, r22, r23, r24, r25, Z, T(SREG)
*/
.global SD_Cue_Sector_Nr_Nc
SD_Cue_Sector_Nr_Nc:

	; Load parameters

	movw  ZL,      r24
	ld    r25,     Z       ; Flags
	bst   r25,     4       ; CRC checking state into T
	rcall sdlib_get_secbuf_Z

	; Transform sector address to byte address (SDSC)

	sbrs  r25,     1       ; SDHC card: Sector address as-is.
	rcall sdlib_convsec_sc ; SDSC cards use byte address

	; Prepare SD Read block command (CMD17)

	ldi   r24,     17      ; CMD17, parameter is OK in r23:r22:r21:r20
	call  SDC_Command_Nc
	cpi   r24,     0x00    ; R1 is Ready?
	breq  .+2
	rjmp  sdlib_ret_fl_02r

	; Return, because CMD17 has been sent
	ret


/*
** Completes a cued single sector read without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          3: Timed out during waiting for data token
** Clobbers:
** r0, r20, r21, r22, r23, r24, r25, Z, T(SREG)
*/
.global SD_CompleteCued_Sector_Nr_Nc
SD_CompleteCued_Sector_Nr_Nc:
	; Load parameters

	movw  ZL,      r24
	ld    r25,     Z       ; Flags
	bst   r25,     4       ; CRC checking state into T
	rcall sdlib_get_secbuf_Z

	; Wait for Data Token & Check

	call  SDC_Wait_FF
	cpi   r24,     0xFE    ; Data token OK
	breq  .+2
	rjmp  sdlib_ret_fl_03r

	; Data is ready to be read. Read it.

	rcall sdlib_wait_spi_with_FF
	movw  r20,     ZL
	subi  r21,     0xFE    ; End of target (+ 512 bytes)
	ldi   r24,     0xFF    ; Filler for the SD card
SD_CompleteCued_Sector_l:
	in    r22,     SPI_DR
	out   SPI_DR,  r24
	st    Z+,      r22     ; Store byte
	cp    ZL,      r20
	cpc   ZH,      r21
	in    r0,      SPI_SR  ; Wait for SPI transaction
	sbrs  r0,      SPIF
	rjmp  .-6
	brcs  SD_CompleteCued_Sector_l

	; 512 bytes in, read the CRC

	in    r21,     SPI_DR
	rcall sdlib_wait_spi_with_FF
	in    r20,     SPI_DR

	; Done, correct read

	jmp   SDC_Release      ; Success



/*
** Performs a single sector read with a retry without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** r23:r22: 512b sector address, high
** r21:r20: 512b sector address, low (together they are a proper C uint32)
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          2: CMD17 failed
**          3: Timed out during waiting for data token
** Clobbers:
** r0, r18, r19, r20, r21, r22, r23, r24, r25, X, Z, T(SREG)
*/
.global SDC_Read_Sector_Nc
SDC_Read_Sector_Nc:

	movw  r18,     r20
	movw  XL,      r22
	push  r24
	push  r25
	rcall SD_Read_Sector_Nr_Nc
	pop   r25
	pop   r0
	movw  r22,     XL
	movw  r20,     r18
	cpi   r24,     0x00
	brne  .+2
	ret                    ; Read successful
	mov   r24,     r0      ; Fall through to SD_Read_Sector

/*
** Performs a single sector read without CRC check on data.
**
** Inputs:
** r25:r24: Pointer to SD data structure
** r23:r22: 512b sector address, high
** r21:r20: 512b sector address, low (together they are a proper C uint32)
** Outputs:
**     r24: Zero if operation succeeded. Otherwise one of the followings:
**          2: CMD17 failed
**          3: Timed out during waiting for data token
** Clobbers:
** r0, r20, r21, r22, r23, r24, r25, Z, T(SREG)
*/
.global SD_Read_Sector_Nr_Nc
SD_Read_Sector_Nr_Nc:

	; Load parameters

	movw  ZL,      r24
	ld    r25,     Z       ; Flags
	bst   r25,     4       ; CRC checking state into T
	rcall sdlib_get_secbuf_Z

	; Transform sector address to byte address (SDSC)

	sbrs  r25,     1       ; SDHC card: Sector address as-is.
	rcall sdlib_convsec_sc ; SDSC cards use byte address

	; Prepare SD Read block command (CMD17)

	ldi   r24,     17      ; CMD17, parameter is OK in r23:r22:r21:r20
	call  SDC_Command_Nc
	cpi   r24,     0x00    ; R1 is Ready?
	breq  .+2
	rjmp  sdlib_ret_fl_02r

	; Wait for Data Token & Check

	call  SDC_Wait_FF
	cpi   r24,     0xFE    ; Data token OK
	breq  .+2
	rjmp  sdlib_ret_fl_03r

	; Data is ready to be read. Read it.

	rcall sdlib_wait_spi_with_FF
	movw  r20,     ZL
	subi  r21,     0xFE    ; End of target (+ 512 bytes)
	ldi   r24,     0xFF    ; Filler for the SD card
SD_Read_Sector_l:
	in    r22,     SPI_DR
	out   SPI_DR,  r24
	st    Z+,      r22     ; Store byte
	cp    ZL,      r20
	cpc   ZH,      r21
	in    r0,      SPI_SR  ; Wait for SPI transaction
	sbrs  r0,      SPIF
	rjmp  .-6
	brcs  SD_Read_Sector_l

	; 512 bytes in, read the CRC

	in    r21,     SPI_DR
	rcall sdlib_wait_spi_with_FF
	in    r20,     SPI_DR

	; Done, correct read

	jmp   SDC_Release      ; Success

/*
** Common return blocks
*/
sdlib_ret_fl_02r:
	call  SDC_Release
	ldi   r24,     0x02
	rjmp  sdlib_ret_fl_c
sdlib_ret_fl_03r:
	call  SDC_Release
	ldi   r24,     0x03
sdlib_ret_fl_c:
	ldi   r25,     0x00
	ret



/*
** Internal function to send 0xFF byte and wait for completion. Useful for
** reading data.
**
** Clobbers:
** r0
*/
sdlib_wait_spi_with_FF:

	sec
	in    r0,      SPI_SR  ; Make sure SPIF is ready for polling
	sbc   r0,      r0      ; r0 = 0xFF
	out   SPI_DR,  r0      ; Fall through to sdlib_wait_spi



/*
** Internal function to wait for SPI transaction completion.
**
** Clobbers:
** r0
*/
sdlib_wait_spi:

	in    r0,      SPI_SR
	sbrc  r0,      SPIF
	ret
	rjmp  .-8



/*
** Internal function to fetch sector buffer from SD data structure
**
** Inputs:
** r11:r10: SD data structure
** Outputs:
** ZH: ZL:  Sector buffer address
** Clobbers:
** r0
*/
sdlib_get_secbuf:
	movw  ZL,      r10
sdlib_get_secbuf_Z:
	ldd   r0,      Z + 1
	ldd   ZH,      Z + 2
	mov   ZL,      r0
	ret



/*
** Convert sector address in r23:r22:r21:r20 for SDSC card
*/
sdlib_convsec_sc:

	mov   r23,     r22
	mov   r22,     r21
	mov   r21,     r20
	ldi   r20,     0x00
	lsl   r21
	rol   r22
	rol   r23              ; SDSC cards use byte address
	ret
