/*

  main.c

  Copyright 2018 Matthew T. Pandina. All rights reserved.

  This file is part of the Uzebox Mode 3 with Scrolling Guide:
  https://github.com/artcfox/uzebox-mode-3-with-scrolling-guide

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdint.h>
#include <stdbool.h>
//#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
//#include <avr/pgmspace.h>
//#include <uzebox.h>
#include <stdio.h>

#define PROGMEM
#define VRAM_TILES_H 32
#define VRAM_TILES_V 32
#define SCREEN_TILES_H 28
#define SCREEN_TILES_V 28
#define TILE_WIDTH 8
#define TILE_HEIGHT 8

#define LO8(x) ((uint8_t)((x) & 0xFF))
#define HI8(x) ((uint8_t)(((x) >> 8) & 0xFF))

typedef struct {
  uint8_t scrollX;
  uint8_t scrollY;
} __attribute__ ((packed)) SCREEN;

uint8_t vram[VRAM_TILES_V * VRAM_TILES_H];
SCREEN Screen;

static inline char pgm_read_byte(const char *d)
{
  return *d;
}

static inline void SetTile(uint8_t x, uint8_t y, uint8_t value)
{
  printf("SetTile(%hhu, %hhu, %hhu)\n", x, y, value);
  vram[y * VRAM_TILES_H + x] = value;
}

static inline uint8_t GetTile(uint8_t x, uint8_t y)
{
  printf("GetTile(%hhu, %hhu) = %hhu\n", x, y, vram[y * VRAM_TILES_H + x]);
  return vram[y * VRAM_TILES_H + x];
}

static inline void ClearVram(void)
{
  for (uint16_t i = 0; i < VRAM_TILES_H * VRAM_TILES_V; ++i)
    vram[i] = 0;
}

#include "../data/tileset.inc"

typedef struct {
  uint16_t width;
  uint16_t height;
  uint16_t offset;
  const char *data;
} __attribute__ ((packed)) LEVEL;

void Level_initFromMap(LEVEL *l, const char *map)
{
  l->width = pgm_read_byte(&map[0]);
  l->height = pgm_read_byte(&map[1]);
  l->offset = 2;
  l->data = map;
}

static inline uint8_t Level_getTileAt(LEVEL *l, uint16_t index)
{
  return pgm_read_byte(&l->data[l->offset + index]);
}

void Level_readColumn(LEVEL *l, uint8_t x, uint16_t y, uint8_t *col)
{
  memset(col, -1, VRAM_TILES_V);

  uint8_t tx = x % VRAM_TILES_H;

  for (uint8_t i = 0; i < VRAM_TILES_V - 2; ++i) {
    if ((y + i) > l->height - 1)
      break;
    col[i] = GetTile(tx, (y + i) % VRAM_TILES_V);
  }
  if (y > 1)
    col[30] = GetTile(tx, (y + 30) % VRAM_TILES_V);
  if (y > 0)
    col[31] = GetTile(tx, (y + 31) % VRAM_TILES_V);
}

void Level_drawColumn(LEVEL *l, uint8_t x, uint16_t y, int16_t realX)
{
  printf("Level_drawColumn(%u, %u, %d)\n", x, y, realX);
  if (realX < 0 || (uint16_t)realX > l->width - 1)
    return;

  uint8_t tx = x % VRAM_TILES_H;

  for (uint8_t i = 0; i < VRAM_TILES_V - 2; ++i) {
    if ((y + i) > l->height - 1)
      break;
    uint16_t index = (y + i) * l->width + realX;
    SetTile(tx, (y + i) % VRAM_TILES_V, Level_getTileAt(l, index));
  }
  if (y > 1)
    SetTile(tx, (y + 30) % VRAM_TILES_V,
	    Level_getTileAt(l, ((int16_t)y - 2) * l->width + realX));
  if (y > 0)
    SetTile(tx, (y + 31) % VRAM_TILES_V,
	    Level_getTileAt(l, ((int16_t)y - 1) * l->width + realX));
}

void Level_readRow(LEVEL *l, uint16_t x, uint8_t y, uint8_t *row)
{
  memset(row, -1, VRAM_TILES_H);

  uint8_t ty = y % VRAM_TILES_V;

  for (uint8_t i = 0; i < VRAM_TILES_H - 2; ++i) {
    if ((x + i) > l->width - 1)
      break;
    row[i] = GetTile((x + i) % VRAM_TILES_H, ty);
  }
  if (x > 1)
    row[30] = GetTile((x + 30) % VRAM_TILES_H, ty);
  if (x > 0)
    row[31] = GetTile((x + 31) % VRAM_TILES_H, ty);
}

void Level_drawRow(LEVEL *l, uint16_t x, uint8_t y, int16_t realY)
{
  if (realY < 0 || (uint16_t)realY > l->height - 1)
    return;

  uint8_t ty = y % VRAM_TILES_V;

  for (uint8_t i = 0; i < VRAM_TILES_H - 2; ++i) {
    if ((x + i) > l->width - 1)
      break;
    uint16_t index = realY * l->width + (x + i);
    SetTile((x + i) % VRAM_TILES_H, ty, Level_getTileAt(l, index));
  }
  if (x > 1)
    SetTile((x + 30) % VRAM_TILES_H, ty,
	    Level_getTileAt(l, realY * l->width + ((int16_t)x - 2)));
  if (x > 0)
    SetTile((x + 31) % VRAM_TILES_H, ty,
	    Level_getTileAt(l, realY * l->width + ((int16_t)x - 1)));
}

typedef struct {
  int16_t x;
  int16_t y;
  LEVEL *level;
} __attribute__ ((packed)) CAMERA;

void Camera_init(CAMERA *c, LEVEL *l)
{
  c->level = l;
  c->x = c->y = 0;
  Screen.scrollX = (uint8_t)c->x;
  Screen.scrollY = (uint8_t)c->y;
}

void Camera_moveTo(CAMERA *c, int16_t x, int16_t y)
{
  if (x < 0) {
    c->x = 0;
  } else {
    int16_t xMax = (c->level->width - SCREEN_TILES_H) * TILE_WIDTH;
    if (x > xMax)
      c->x = xMax;
    else
      c->x = x;
  }

  if (y < 0) {
    c->y = 0;
  } else {
    int16_t yMax = (c->level->height - SCREEN_TILES_V) * TILE_HEIGHT;
    if (y > yMax)
      c->y = yMax;
    else
      c->y = y;
  }
}

void Camera_fillVram(CAMERA *c)
{
  int16_t cxt = c->x / TILE_WIDTH;
  int16_t cyt = c->y / TILE_HEIGHT;

  for (uint8_t i = 0; i < VRAM_TILES_H - 2; ++i)
    Level_drawColumn(c->level, cxt + i, cyt, cxt + i);

  Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2);
  Level_drawColumn(c->level, cxt + 31, cyt, cxt - 1);
}

void Camera_readVram(CAMERA *c, uint8_t *buf)
{
  int16_t cxt = c->x / TILE_WIDTH;
  int16_t cyt = c->y / TILE_HEIGHT;

  uint8_t col = 0;
  for (uint8_t i = 0; i < VRAM_TILES_H - 2; ++i) {
    Level_readColumn(c->level, cxt + i, cyt, buf + col++ * VRAM_TILES_V);
    //Level_drawColumn(c->level, cxt + i, cyt, cxt + i);
  }
  Level_readColumn(c->level, cxt + 30, cyt, buf + col++ * VRAM_TILES_V);
  Level_readColumn(c->level, cxt + 31, cyt, buf + col++ * VRAM_TILES_V);
  //Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2);
  //Level_drawColumn(c->level, cxt + 31, cyt, cxt - 1);
}

void Camera_update(CAMERA *c, uint8_t *buf)
{
  /* uint8_t prevX = Screen.scrollX; */
  /* uint8_t prevY = Screen.scrollY; */
  /* Screen.scrollX = (uint8_t)c->x; */
  /* Screen.scrollY = (uint8_t)c->y; */

  int16_t cxt = c->x / TILE_WIDTH;
  int16_t cyt = c->y / TILE_HEIGHT;

  // Have we scrolled past a tile boundary along X?
  /* if ((prevX & ~(TILE_WIDTH - 1)) != (Screen.scrollX & ~(TILE_WIDTH - 1))) { */
  /*   if ((uint8_t)(Screen.scrollX - prevX) < (uint8_t)(prevX - Screen.scrollX)) */
      Level_drawColumn(c->level, cxt + 29, cyt, cxt + 29);
      Level_readColumn(c->level, cxt + 29, cyt, buf);

      for (uint8_t i = 0; i < VRAM_TILES_V; ++i)
	printf("%03hhu ", buf[i]);
      printf("\n");

    /* else */
      Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2);
      Level_readColumn(c->level, cxt + 30, cyt, buf + VRAM_TILES_V);

      for (uint8_t i = 0; i < VRAM_TILES_V; ++i)
	printf("%03hhu ", buf[VRAM_TILES_V + i]);
      printf("\n");


  /* } */
  // Have we scrolled past a tile boundary along Y?
  /* if ((prevY & ~(TILE_HEIGHT - 1)) != (Screen.scrollY & ~(TILE_HEIGHT - 1))) { */
  /*   if ((uint8_t)(Screen.scrollY - prevY) < (uint8_t)(prevY - Screen.scrollY)) */
      Level_drawRow(c->level, cxt, cyt + 29, cyt + 29);
      Level_readRow(c->level, cxt, cyt + 29, buf + VRAM_TILES_V * 2);

      for (uint8_t i = 0; i < VRAM_TILES_H; ++i)
	printf("%03hhu ", buf[VRAM_TILES_V * 2 + i]);
      printf("\n");

    /* else */
      Level_drawRow(c->level, cxt, cyt + 30, cyt - 2);
      Level_readRow(c->level, cxt, cyt + 30, buf + VRAM_TILES_V * 2 + VRAM_TILES_H);

      for (uint8_t i = 0; i < VRAM_TILES_H; ++i)
	printf("%03hhu ", buf[VRAM_TILES_V * 2 + VRAM_TILES_H + i]);
      printf("\n");

  /* } */
}

void PrintVram(CAMERA *c)
{
  printf("Camera: (%d, %d)\n", c->x, c->y);
  for (uint8_t y = 0; y < VRAM_TILES_V; ++y) {
    for (uint8_t x = 0; x < VRAM_TILES_H; ++x) {
      printf("%02hu, ", vram[y * VRAM_TILES_H + x]);
    }
    putchar('\n');
  }
  putchar('\n');
}

int main()
{
  //SetTileTable(tileset);

  LEVEL l;
  CAMERA c;

  ClearVram();

  Level_initFromMap(&l, map_level);
  Camera_init(&c, &l);
  //Camera_fillVram(&c);

  /* int16_t cxt = c.x / TILE_WIDTH; */
  /* int16_t cyt = c.y / TILE_HEIGHT; */

  // For every valid camera position (x, y) create an index:
  //   index = y * (level->width - SCREEN_TILES_H + 1) + x
  // That index becomes the index to the sector (possibly packed)
  // containing each of the possible forward/backward level strips
  // for that camera position.

  // The idea is that we're going to loop over every valid camera
  // position (on a tile boundary), call the Camera_moveTo function
  // to set the camera position, draw all four strips into VRAM, and
  // then read them back out in a defined order, writing them to the
  // appropriatly indexed sector in the output file.

  // Every time we move the camera, we're going to call ClearVram()
  // (or an equivalent function) to populate VRAM with dummy values
  // in the case where Level_drawColumn and/or Level_drawRow skips
  // writing to VRAM, because it's outside of the level bounds. That
  // dummy value should be something that can help us debug things.

  printf("Level Dimensions: %u, %u\n", l.width, l.height);

  uint16_t initCameraX = 0 + 0;
  uint16_t initCameraY = 0 + 0;
  uint16_t initPlayerX = 4 + 5;
  uint16_t initPlayerY = 4 + 7;

  FILE *fd = fopen("sdlevel1.dat", "w");
  uint8_t header[512];
  memset(header, 0, 512);
  header[0] = LO8(l.width); // width
  header[1] = HI8(l.width);
  header[2] = LO8(l.height); // height
  header[3] = HI8(l.height);
  header[4] = LO8(initCameraX);
  header[5] = HI8(initCameraX);
  header[6] = LO8(initCameraY);
  header[7] = HI8(initCameraY);
  header[8] = LO8(initPlayerX);
  header[9] = HI8(initPlayerX);
  header[10] = LO8(initPlayerY);
  header[11] = HI8(initPlayerY);
  header[12] = 3; // sector where hidden level strip data begins
  fwrite(header, sizeof(uint8_t), 512, fd);

  uint8_t initVram[1024];
  memset(initVram, 0, 1024);
  ClearVram();
  Camera_moveTo(&c, initCameraX, initCameraY);
  Camera_fillVram(&c);
  Camera_readVram(&c, initVram); // XXX
  fwrite(initVram, sizeof(uint8_t), 1024, fd);

  for (uint16_t ty = 0; ty < l.height - SCREEN_TILES_V + 1; ++ty) {
    for (uint16_t tx = 0; tx < l.width - SCREEN_TILES_H + 1; ++tx) {
      printf("tx: %02x, ty: %02x, sector %04x\n", tx, ty, ty * (l.width - SCREEN_TILES_H + 1) + tx);
      ClearVram();
      Camera_moveTo(&c, tx * TILE_WIDTH, ty * TILE_HEIGHT);

      uint8_t buf[512];
      memset(buf, 0xaa, 512);

      Camera_update(&c, buf);

      fwrite(buf, sizeof(uint8_t), 512, fd);

      // Print the contents of buf
      for (uint16_t i = 0; i < VRAM_TILES_H * 2 + VRAM_TILES_V * 2;) {
	printf("%02x ", buf[i]);
	if ((++i % 8) == 0)
	  printf(" ");
	if ((i % 16) == 0)
	  printf("\n");
      }

    }
  }
  fclose(fd);

  /* for (;;) { */
  /*   uint16_t held = ReadJoypad(0); */
  /*   if (held & BTN_LEFT) */
  /*     Camera_moveTo(&c, c.x - 1, c.y); */
  /*   else if (held & BTN_RIGHT) */
  /*     Camera_moveTo(&c, c.x + 1, c.y); */

  /*   if (held & BTN_UP) */
  /*     Camera_moveTo(&c, c.x, c.y - 1); */
  /*   else if (held & BTN_DOWN) */
  /*     Camera_moveTo(&c, c.x, c.y + 1); */

  /*   Camera_update(&c); */

  /*   WaitVsync(1); */
  /* } */
}
