/*

  sdmap.c

  Copyright 2018 Matthew T. Pandina. All rights reserved.

  This is free software: you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with this.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdint.h>
#include <stdbool.h>
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <uzebox.h>
#include <bootlib.h>
#include "fastread.h"

#include "data/tileset.inc"
#include "data/sprites.inc"
#include "data/patches.inc"

#define UZEMCHR _SFR_IO8(26) //uzem whisper port for outputting characters to the console
#define UZEMHEX _SFR_IO8(25) //uzem whisper port for outputting hex bytes values to the console

typedef struct {
  uint16_t held;
  uint16_t prev;
  uint16_t pressed;
  uint16_t released;
} __attribute__ ((packed)) BUTTON_INFO;

typedef struct {
  int16_t x;
  int16_t y;
  int16_t dx;
  int16_t dy;
  int16_t maxdx;
  int16_t impulse;
  uint8_t animationFrameCounter;
  uint8_t framesFalling;
  unsigned int interacts:1;
  unsigned int falling:1;
  unsigned int jumping:1;
  unsigned int left:1;
  unsigned int right:1;
  unsigned int up:1;
  unsigned int down:1;
  unsigned int jump:1;
  unsigned int jumpReleased:1;
  unsigned int turbo:1;
  unsigned int attackReleased:1;
  unsigned int visible:1;
  unsigned int dead:1;
  unsigned int attacking:1;
  unsigned int invincible:1;
  BUTTON_INFO buttons;
} __attribute__ ((packed)) PLAYER;

#define TILE_END_SKY         0
#define TILE_START_ONE_WAY   1
#define TILE_END_ONE_WAY     4
#define TILE_START_SOLID     5

#define isSolid(t) ((t) >= TILE_START_SOLID)
#define isOneWay(t) (((t) >= TILE_START_ONE_WAY) && ((t) <= TILE_END_ONE_WAY))
//#define isLadder(t) ((((t) >= FIRST_UNDERGROUND_LADDER_TOP_TILE) && ((t) <= LAST_UNDERGROUND_LADDER_MIDDLE_TILE)) || (((t) >= FIRST_ABOVEGROUND_ONE_WAY_LADDER_TOP_TILE) && ((t) <= LAST_SKY_LADDER_MIDDLE_TILE)))

#define NELEMS(x) (sizeof(x)/sizeof(x[0]))

// Gives a compile-time error if condition is not satisfied, produces no code otherwise
#define BUILD_BUG_ON(condition) ((void)sizeof(char[1 - 2*!!(condition)]))
#define isNotPowerOf2(x) (((x) & ((x) - 1)) != 0)

#define vt2p(t) ((t) * (TILE_HEIGHT << FP_SHIFT))
#define ht2p(t) ((t) * (TILE_WIDTH << FP_SHIFT))
#define p2vt(p) ((p) / (TILE_HEIGHT << FP_SHIFT))
#define p2ht(p) ((p) / (TILE_WIDTH << FP_SHIFT))
#define nv(p) ((p) % (TILE_HEIGHT << FP_SHIFT))
#define nh(p) ((p) % (TILE_WIDTH << FP_SHIFT))

#define nearestScreenPixel(p) (((p) + (1 << (FP_SHIFT - 1))) >> FP_SHIFT)

// The contents of these arrays are automatically generated based on a set
// of .xcf files inside the data directory. To add levels, add them to the
// Makefile and to the PROGMEM arrays below.

const uint16_t levelSize[] PROGMEM = {
  60, 56
};

#define NUM_LEVELS (NELEMS(levelSize) / 2)

#define FP_SHIFT 2
#define WORLD_FPS 32
#define WORLD_METER (10 << FP_SHIFT)
#define WORLD_GRAVITY 615
#define WORLD_MAXDX (WORLD_METER * 6)
#define WORLD_MAXDY (WORLD_METER * 16)
#define WORLD_ACCEL 355
#define WORLD_FRICTION 437
#define WORLD_JUMP 17600
#define WORLD_FALLING_GRACE_FRAMES 6
#define WORLD_CUT_JUMP_SPEED_LIMIT (WORLD_GRAVITY / 10)

#define LO8(x) ((uint8_t)((x) & 0xFF))
#define HI8(x) ((uint8_t)(((x) >> 8) & 0xFF))

typedef struct {
  uint16_t width;
  uint16_t height;
  uint16_t initCameraX;
  uint16_t initCameraY;
  uint16_t initPlayerX;
  uint16_t initPlayerY;
  uint16_t offset;

  sdc_struct_t sds;
  uint8_t buf[512];
  uint32_t startSector;
  //  const char *data;
} __attribute__ ((packed)) LEVEL;

/* void Level_initFromMap(LEVEL *l, const char *map) */
/* { */
/*   l->width = pgm_read_byte(&map[0]); */
/*   l->height = pgm_read_byte(&map[1]); */
/*   l->offset = 2; */
/*   l->data = map; */
/* } */

void Level_initFromSD(LEVEL *l)
{
  // Find start sector for PROGMEM filename string
  // Read header (stored in its first sector)
  // width, height, initial camera x, initial camera y,
  // initial vram offset, strips offset,
  // ---- new sector
  // rows of initial vram data
  // ... (spans 2 sectors)
  // ---- new sector
  // strip data, +X, -X, +Y, -Y
  uint32_t cluster;

  l->sds.bufp = &l->buf[0];
  uint8_t res = FS_Init(&l->sds);
  if (res != 0U)
    while (1);

  cluster = FS_Find(&l->sds,
	    ((u16)('S') << 8) |
	    ((u16)('D')     ),
	    ((u16)('L') << 8) |
	    ((u16)('E')     ),
	    ((u16)('V') << 8) |
	    ((u16)('E')     ),
	    ((u16)('L') << 8) |
	    ((u16)('1')     ),
	    ((u16)('D') << 8) |
	    ((u16)('A')     ),
	    ((u16)('T') << 8) |
	    ((u16)(0)       ));
  if (cluster == 0U)
    while (1);
  FS_Select_Cluster(&l->sds, cluster);
  l->startSector = FS_Get_Sector(&l->sds);
  SPI_Set_Max();
  SDC_CRC_Enable(&l->sds, 0);
  SD_Read_Sector_Nr_Nc(&l->sds, l->startSector);
  l->width = l->buf[0] | (l->buf[1] << 8);
  l->height = l->buf[2] | (l->buf[3] << 8);
  l->initCameraX = l->buf[4] | (l->buf[5] << 8);
  l->initCameraY = l->buf[6] | (l->buf[7] << 8);
  l->initPlayerX = l->buf[8] | (l->buf[9] << 8);
  /* UZEMHEX = l->buf[9]; UZEMHEX = l->buf[8]; UZEMCHR = '\n'; */
  l->initPlayerY = l->buf[10] | (l->buf[11] << 8);
  /* UZEMHEX = l->buf[11]; UZEMHEX = l->buf[10]; UZEMCHR = '\n'; */
  //l->width = 60; l->height = 56;
  l->offset = l->buf[12];
}

/* static inline uint8_t Level_getTileAt(LEVEL *l, uint16_t index) */
/* { */
/*   return pgm_read_byte(&l->data[l->offset + index]); */
/* } */

/* void Level_drawColumn(LEVEL *l, uint8_t x, uint16_t y, int16_t realX) */
/* { */
/*   if (realX < 0 || (uint16_t)realX > l->width - 1) */
/*     return; */

/*   uint8_t tx = x % VRAM_TILES_H; */

/*   for (uint8_t i = 0; i < VRAM_TILES_V - 2; ++i) { */
/*     if ((y + i) > l->height - 1) */
/*       break; */
/*     uint16_t index = (y + i) * l->width + realX; */
/*     SetTile(tx, (y + i) % VRAM_TILES_V, Level_getTileAt(l, index)); */
/*   } */
/*   if (y > 1) */
/*     SetTile(tx, (y + 30) % VRAM_TILES_V, */
/* 	    Level_getTileAt(l, ((int16_t)y - 2) * l->width + realX)); */
/*   if (y > 0) */
/*     SetTile(tx, (y + 31) % VRAM_TILES_V, */
/* 	    Level_getTileAt(l, ((int16_t)y - 1) * l->width + realX)); */
/* } */

/* void Level_drawRow(LEVEL *l, uint16_t x, uint8_t y, int16_t realY) */
/* { */
/*   if (realY < 0 || (uint16_t)realY > l->height - 1) */
/*     return; */

/*   uint8_t ty = y % VRAM_TILES_V; */

/*   for (uint8_t i = 0; i < VRAM_TILES_H - 2; ++i) { */
/*     if ((x + i) > l->width - 1) */
/*       break; */
/*     uint16_t index = realY * l->width + (x + i); */
/*     SetTile((x + i) % VRAM_TILES_H, ty, Level_getTileAt(l, index)); */
/*   } */
/*   if (x > 1) */
/*     SetTile((x + 30) % VRAM_TILES_H, ty, */
/* 	    Level_getTileAt(l, realY * l->width + ((int16_t)x - 2))); */
/*   if (x > 0) */
/*     SetTile((x + 31) % VRAM_TILES_H, ty, */
/* 	    Level_getTileAt(l, realY * l->width + ((int16_t)x - 1))); */
/* } */

typedef struct {
  int16_t x;
  int16_t y;
  int16_t prevX;
  int16_t prevY;
  LEVEL *level;

  bool needColumn;
  bool needRow;
  bool forwardX;
  bool forwardY;
} __attribute__ ((packed)) CAMERA;

void Camera_init(CAMERA *c, LEVEL *l)
{
  c->level = l;
  //c->x = c->y = 0;
  c->x = l->initCameraX;
  c->y = l->initCameraY;
  Screen.scrollX = (uint8_t)c->x;
  Screen.scrollY = (uint8_t)c->y;
  c->needColumn = false;
  c->needRow = false;

}

void Camera_moveTo(CAMERA *c, int16_t x, int16_t y)
{
  c->prevX = x;
  c->prevY = y;

  if (x < 0) {
    c->x = 0;
  } else {
    int16_t xMax = (c->level->width - SCREEN_TILES_H) * TILE_WIDTH;
    if (x > xMax)
      c->x = xMax;
    else
      c->x = x;
  }

  if (y < 0) {
    c->y = 0;
  } else {
    int16_t yMax = (c->level->height - SCREEN_TILES_V) * TILE_HEIGHT;
    if (y > yMax)
      c->y = yMax;
    else
      c->y = y;
  }
}

void Camera_fillVramFromSD(CAMERA *c)
{
  int16_t cxt = c->x / TILE_WIDTH;
  int16_t cyt = c->y / TILE_HEIGHT;

#define HEADER_SIZE 1
  uint16_t calculatedSectorOffset = HEADER_SIZE;
  SD_Read_Sector_Nr_Nc(&c->level->sds, c->level->startSector + calculatedSectorOffset);

  uint8_t col = 0;
  for (uint8_t i = 0; i < 16; ++i) {
    uint8_t tx = (cxt + i) % VRAM_TILES_H;
    for (uint8_t j = 0; j < VRAM_TILES_V; ++j) {
      SetTile(tx, ((uint16_t)cyt + j) % VRAM_TILES_V, c->level->buf[col * VRAM_TILES_V + j]);
    }
    col++;
    //Level_drawColumn(c->level, cxt + i, cyt, cxt + i);
  }

  SD_Read_Sector_Nr_Nc(&c->level->sds, c->level->startSector + calculatedSectorOffset + 1);
  col = 0;

  for (uint8_t i = 16; i < VRAM_TILES_H - 2; ++i) {
    uint8_t tx = (cxt + i) % VRAM_TILES_H;
    for (uint8_t j = 0; j < VRAM_TILES_V; ++j) {
      SetTile(tx, ((uint16_t)cyt + j) % VRAM_TILES_V, c->level->buf[col * VRAM_TILES_V + j]);
    }
    col++;
    //Level_drawColumn(c->level, cxt + i, cyt, cxt + i);
  }

  uint8_t tx = (cxt + 30) % VRAM_TILES_H;
  for (uint8_t j = 0; j < VRAM_TILES_V; ++j) {
    SetTile(tx, ((uint16_t)cyt + j) % VRAM_TILES_V, c->level->buf[col * VRAM_TILES_V + j]);
  }
  col++;
  //Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2);

  tx = (cxt + 31) % VRAM_TILES_H;
  for (uint8_t j = 0; j < VRAM_TILES_V; ++j) {
    SetTile(tx, ((uint16_t)cyt + j) % VRAM_TILES_V, c->level->buf[col * VRAM_TILES_V + j]);
  }
  col++;
  //Level_drawColumn(c->level, cxt + 31, cyt, cxt - 1);
}

/* void Camera_fillVram(CAMERA *c) */
/* { */
/*   int16_t cxt = c->x / TILE_WIDTH; */
/*   int16_t cyt = c->y / TILE_HEIGHT; */

/*   for (uint8_t i = 0; i < VRAM_TILES_H - 2; ++i) */
/*     Level_drawColumn(c->level, cxt + i, cyt, cxt + i); */

/*   Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2); */
/*   Level_drawColumn(c->level, cxt + 31, cyt, cxt - 1); */
/* } */

void Camera_update(CAMERA *c)
{
  uint8_t prevX = Screen.scrollX;
  uint8_t prevY = Screen.scrollY;
  Screen.scrollX = (uint8_t)c->x;
  Screen.scrollY = (uint8_t)c->y;

  int16_t cxt = c->x / TILE_WIDTH;
  int16_t cyt = c->y / TILE_HEIGHT;

  bool needColumn = ((prevX & ~(TILE_WIDTH - 1)) != (Screen.scrollX & ~(TILE_WIDTH - 1)));
  bool needRow = ((prevY & ~(TILE_HEIGHT - 1)) != (Screen.scrollY & ~(TILE_HEIGHT - 1)));
  //const uint16_t byteOffset = 0;
  if (needColumn || needRow) {
    /* uint16_t calculatedSectorOffset = (cyt * c->level->width + cxt) / 4; */
    /* byteOffset = ((cyt * c->level->width + cxt) % 4) * (512 / 4); */
    uint16_t calculatedSectorOffset = c->level->offset + (cyt * (c->level->width - SCREEN_TILES_H + 1) + cxt);
    SD_Cue_Sector_Nr_Nc(&c->level->sds, c->level->startSector + calculatedSectorOffset);

    UZEMCHR = 'x'; UZEMCHR = ' '; UZEMHEX = HI8((uint16_t)cxt); UZEMHEX = LO8((uint16_t)cxt); UZEMCHR = ','; UZEMCHR = ' ';
    UZEMCHR = 'y'; UZEMCHR = ' '; UZEMHEX = HI8((uint16_t)cyt); UZEMHEX = LO8((uint16_t)cyt); UZEMCHR = ','; UZEMCHR = ' ';
    UZEMCHR = 's'; UZEMCHR = ' '; UZEMHEX = HI8(calculatedSectorOffset); UZEMHEX = LO8(calculatedSectorOffset); UZEMCHR = '\n';

    /* for (uint16_t j = 0; j < 512;) { */
    /*   UZEMHEX = c->level->sds.bufp[j]; */
    /*   UZEMCHR = ' '; */
    /*   if ((++j % 8) == 0) */
    /* 	UZEMCHR = ' '; */
    /*   if ((j % 16) == 0) */
    /* 	UZEMCHR = '\n'; */
    /* } */

  }
  c->needColumn = needColumn;
  c->needRow = needRow;
  c->forwardX = ((uint8_t)(Screen.scrollX - prevX) < (uint8_t)(prevX - Screen.scrollX));
  c->forwardY = ((uint8_t)(Screen.scrollY - prevY) < (uint8_t)(prevY - Screen.scrollY));
  #if 0
  const uint16_t byteOffset = 0;
  // Have we scrolled past a tile boundary along X?
  if (needColumn) {
    if ((uint8_t)(Screen.scrollX - prevX) < (uint8_t)(prevX - Screen.scrollX)) {
      // I think I might need to look at the rows/cols based on the previous camera position!
      uint8_t tx = (cxt + 29) % VRAM_TILES_H;
      for (uint8_t i = 0; i < VRAM_TILES_V; ++i) {
	SetTile(tx, ((uint16_t)cyt + i) % VRAM_TILES_V, c->level->buf[byteOffset + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawColumn(c->level, cxt + 29, cyt, cxt + 29);
    } else {
      uint8_t tx = (cxt + 30) % VRAM_TILES_H;
      for (uint8_t i = 0; i < VRAM_TILES_V; ++i) {
	SetTile(tx, ((uint16_t)cyt + i) % VRAM_TILES_V, c->level->buf[byteOffset + 32 + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + 32 + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2);
    }
  }
  // Have we scrolled past a tile boundary along Y?
  if (needRow) {
    if ((uint8_t)(Screen.scrollY - prevY) < (uint8_t)(prevY - Screen.scrollY)) {
      uint8_t ty = (cyt + 29) % VRAM_TILES_V;
      for (uint8_t i = 0; i < VRAM_TILES_H; ++i) {
	SetTile(((uint16_t)cxt + i) % VRAM_TILES_H, ty, c->level->buf[byteOffset + 64 + i]);
	//SetTile(i, ty, c->buf[byteOffset + 64 + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + 64 + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawRow(c->level, cxt, cyt + 29, cyt + 29);
    } else {
      uint8_t ty = (cyt + 30) % VRAM_TILES_V;
      for (uint8_t i = 0; i < VRAM_TILES_H; ++i) {
	SetTile(((uint16_t)cxt + i) % VRAM_TILES_H, ty, c->level->buf[byteOffset + 96 + i]);
	//SetTile(i, ty, c->buf[byteOffset + 64 + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + 64 + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawRow(c->level, cxt, cyt + 30, cyt - 2);
    }
  }
  #endif
}

__attribute__(( always_inline ))
static inline bool isSolidForEntity(const uint8_t tx, const uint8_t ty, const int16_t y, const int16_t prevY, const uint8_t entityHeight, const bool down)
{
  uint8_t t = GetTile(tx, ty);
  //uint8_t = vram[offset] - RAM_TILES_COUNT; // equiv. GetTile(tx, ty)

  // One-way tiles are only solid for Y collisions where your previous Y puts your feet above the tile, and you're not currently pressing down
  return (isSolid(t) || (isOneWay(t) && !down && ((prevY + entityHeight - 1) < vt2p(p2vt(y)))));
}

void Player_init(PLAYER* const p, LEVEL* const l)
{
  memset(p, 0, sizeof(PLAYER));
  /* p->x = 32 << FP_SHIFT; */
  /* p->y = 32 << FP_SHIFT; */
  p->x = (l->initPlayerX * TILE_WIDTH) << FP_SHIFT;
  p->y = (l->initPlayerY * TILE_HEIGHT) << FP_SHIFT;
  p->maxdx = WORLD_MAXDX;
  p->impulse = WORLD_JUMP;
  p->visible = true;
  p->jumpReleased = true;
  p->interacts = true;
  p->right = true;
  p->falling = true;
  p->attacking = false;
  p->attackReleased = true;
}

#define PLAYER_ANIMATION_FRAME_SKIP 8
#define PLAYER_ATTACK_ANIMATION_FRAME_SKIP 4
/* #define PLAYER_LADDER_ANIMATION_FRAME_SKIP 4 */
const uint8_t playerAnimation[] PROGMEM = { 0, 1, 2, 1, 0, 1, 2, 1 };
const uint8_t playerAttackAnimation[] PROGMEM = { 0, 1, 2, 3, 4, 4, 4, 4 };
const VRAM_PTR_TYPE* const walk_maps[] PROGMEM = { map_walk0, map_walk1, map_walk2, map_walk1 };
const VRAM_PTR_TYPE* const run_maps[] PROGMEM = { map_run0, map_run1, map_run2, map_run1 };
const VRAM_PTR_TYPE* const attack_maps[] PROGMEM = { map_attack0, map_attack1, map_attack2, map_attack3, map_attack4 };
//const uint8_t playerLadderAnimation[] PROGMEM = { 0, 1, 2, 1, 0, 3, 4, 3 };

void Player_input(PLAYER* const p)
{
  p->buttons.prev = p->buttons.held;
  p->buttons.held = ReadJoypad(0);
  p->buttons.pressed = p->buttons.held & (p->buttons.held ^ p->buttons.prev);
  p->buttons.released = p->buttons.prev & (p->buttons.held ^ p->buttons.prev);

  if (!p->attacking && p->attackReleased && (p->buttons.held & BTN_SL) && !p->jumping && !p->falling) {
    p->attacking = true;
    p->attackReleased = false;
    p->animationFrameCounter = 0;
  }

  if ((p->buttons.released & BTN_SL))
    p->attackReleased = true;

  if (!p->attacking) {
    p->left = (bool)(p->buttons.held & BTN_LEFT);
    p->right = (bool)(p->buttons.held & BTN_RIGHT);
    p->up = (bool)(p->buttons.held & BTN_UP);
    p->down = (bool)(p->buttons.held & BTN_DOWN);
    p->turbo = (bool)(p->buttons.held & BTN_B);
  }

  // Allow players to jump by holding the jump button before landing,
  // but require them to release it before they can jump again
  if (p->jumpReleased) {
    p->jump = (bool)(p->buttons.held & BTN_A);

    if (p->jump /*&& e->update == Player_update*/) {
      // Look at the tile(s) above the player's head. If the jump would not
      // be allowed, then don't set jump to true until they can actually jump
      // This allows the player to hold the jump button early while
      // jump-restricted, but still make the jump as soon as it is allowed
      int16_t roundedX = nearestScreenPixel(p->x) << FP_SHIFT; // ignore subpixels for this calculation
      uint8_t tx = p2ht(roundedX);
      uint8_t ty = p2vt(p->y - 1);
      if (                       (GetTile(tx % VRAM_TILES_H,       ty % VRAM_TILES_V) >= TILE_START_SOLID) || // cellup,     equiv. ... GetTile(tx,     ty) ...
	  ((bool)nh(roundedX) && (GetTile((tx + 1) % VRAM_TILES_H, ty % VRAM_TILES_V) >= TILE_START_SOLID)))  // cellupdiag, equiv. ... GetTile(tx + 1, ty) ...
        p->jump = false;
    }

  } else {
    p->jump = false;
    if (p->buttons.released & BTN_A)
      p->jumpReleased = true;
  }
}

__attribute__((optimize("O3")))
void Player_update(PLAYER* const p)
{
  bool wasLeft = (p->dx < 0);
  bool wasRight = (p->dx > 0);

  int16_t ddx = 0;

  if (p->left)
    ddx -= WORLD_ACCEL;    // entity wants to go left
  else if (wasLeft)
    ddx += WORLD_FRICTION; // entity was going left, but not anymore

  if (p->right)
    ddx += WORLD_ACCEL;    // entity wants to go right
  else if (wasRight)
    ddx -= WORLD_FRICTION; // entity was going right, but not anymore

  // Compile-time assert that we are working with a power of 2
  BUILD_BUG_ON(isNotPowerOf2(WORLD_FPS));

  // Integrate the X forces to calculate the new position (x,y) and the new velocity (dx,dy)
  p->x += (p->dx / WORLD_FPS);
  p->dx += (ddx / WORLD_FPS);
  if (p->turbo) {
    if (p->dx < -(WORLD_MAXDX + WORLD_METER))
      p->dx = -(WORLD_MAXDX + WORLD_METER);
    else if (p->dx > (WORLD_MAXDX + WORLD_METER))
      p->dx = (WORLD_MAXDX + WORLD_METER);
  } else {
    if (p->dx < -WORLD_MAXDX)
      p->dx = -WORLD_MAXDX;
    else if (p->dx > WORLD_MAXDX)
      p->dx = WORLD_MAXDX;
  }

  // We can't clamp to screen bounds in mode 3 scrolling

  // Clamp X to within screen bounds
  /* if (p->x > ((SCREEN_TILES_H - 1) * (TILE_WIDTH << FP_SHIFT))) { */
  /*   p->x = ((SCREEN_TILES_H - 1) * (TILE_WIDTH << FP_SHIFT)); */
  /*   p->dx = 0; */
  /* } else if (p->x < 0) { */
  /*   p->x = 0; */
  /*   p->dx = 0; */
  /* } */

  // Clamp horizontal velocity to zero if we detect that the direction has changed
  if ((wasLeft && (p->dx > 0)) || (wasRight && (p->dx < 0)))
    p->dx = 0; // clamp at zero to prevent friction from making the entity jiggle side to side

  // Collision Detection for X
  uint8_t tx = p2ht(p->x);
  bool nx = (bool)nh(p->x); // true if entity overlaps right
  uint8_t ty = p2vt(p->y);
  bool ny = (bool)nv(p->y); // true if entity overlaps below

  bool cell      = isSolid(GetTile(tx % VRAM_TILES_H, ty % VRAM_TILES_V));
  bool cellright = isSolid(GetTile((tx + 1) % VRAM_TILES_H, ty % VRAM_TILES_V));
  bool celldown  = isSolid(GetTile(tx % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V));
  bool celldiag  = isSolid(GetTile((tx + 1) % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V));

  // The tests above comparing GetTile() to TILE_RED should be function calls
  // to isSolid() similar to what was done for Bugz (shown below)

  /* uint16_t offset = ty * SCREEN_TILES_H + tx; */
  /* bool cell      = isSolid(vram[offset                     ] - RAM_TILES_COUNT); // equiv. GetTile(tx,     ty    ) */
  /* bool cellright = isSolid(vram[offset + 1                 ] - RAM_TILES_COUNT); // equiv. GetTile(tx + 1, ty    ) */
  /* bool celldown  = isSolid(vram[offset + SCREEN_TILES_H    ] - RAM_TILES_COUNT); // equiv. GetTile(tx,     ty + 1) */
  /* bool celldiag  = isSolid(vram[offset + SCREEN_TILES_H + 1] - RAM_TILES_COUNT); // equiv. GetTile(tx + 1, ty + 1) */

  if (p->dx > 0) {
    if ((nx && cellright && !cell) || // nx check avoids potential glitch when moving off ladder
        (ny && celldiag  && !celldown)) {
      p->x = ht2p(p2ht(p->x)); // clamp the x position to avoid moving into the platform just hit
      p->dx = 0;           // stop horizontal velocity
    }
  } else if (p->dx < 0) {
    if ((nx && cell     && !cellright) || // nx check avoids potential glitch when moving off ladder
        (ny && celldown && !celldiag)) {
      p->x = ht2p(p2ht(p->x) + 1); // clamp the x position to avoid moving into the platform just hit
      p->dx = 0;           // stop horizontal velocity
    }
  }

  int16_t ddy = WORLD_GRAVITY;

  // Not needed for simple demo
#if 0
  // Bounce a bit when you stomp a monster
  if (p->monsterhop) {
    p->monsterhop = p->jumping = p->falling = false;
    p->jumpReleased = true;
    p->framesFalling = 0;
    if (p->dy > 0)
      p->dy = 0;           // if falling down, reset vertical velocity
    ddy -= (WORLD_JUMP >> 1);
  }
#endif

  // Attack/Jump logic
  if (p->attacking) {
    p->dx = 0;
    p->dy = 0;
    // Should probably use a dedicated counter for how long the attack lasts
    if (p->animationFrameCounter == (PLAYER_ATTACK_ANIMATION_FRAME_SKIP * NELEMS(playerAttackAnimation) - 1))
      p->attacking = false;
  } else if (p->jump && !p->jumping && !(p->falling ? (p->framesFalling > WORLD_FALLING_GRACE_FRAMES) : false)) {
    TriggerFx(0, 128, true);
    if (p->dy > 0)
      p->dy = 0;           // if falling down, reset vertical velocity so jumps during grace period are consistent with jumps from ground
    ddy -= WORLD_JUMP;     // apply an instantaneous (large) vertical impulse
    p->jumping = true;
    p->jumpReleased = false;
  }

  // Variable height jumping
  if (p->jumping && p->jumpReleased && (p->dy < -WORLD_CUT_JUMP_SPEED_LIMIT))
      p->dy = -WORLD_CUT_JUMP_SPEED_LIMIT;

  // Integrate the Y forces to calculate the new position (x,y) and the new velocity (dx,dy)
  int16_t prevY = p->y; // cache previous Y value for one-way tiles
  p->y += (p->dy / WORLD_FPS);
  p->dy += (ddy / WORLD_FPS);
  if (p->dy < -WORLD_MAXDY)
    p->dy = -WORLD_MAXDY;
  else if (p->dy > WORLD_MAXDY)
    p->dy = WORLD_MAXDY;

  // We can't clamp to screen bounds in mode 3 scrolling

  // Clamp Y to within screen bounds
  /* if (p->y > vt2p(SCREEN_TILES_V - 1)) { */
  /*   p->y = vt2p(SCREEN_TILES_V - 1); */
  /*   p->dy = 0; */
  /*   // Kill the entity if it would have fallen through the bottom of the screen */
  /*   if (!p->invincible) */
  /*     p->dead = true; */
  /*   return; */
  /* } else if (p->y < 0) { */
  /*   p->y = 0; */
  /*   p->dy = 0; */
  /* } */

  // Collision Detection for Y (uses rounded X so if it looks like the entity should fall through a one-tile-wide hole, it will)
  int16_t roundedX = (p->dx == 0) ? nearestScreenPixel(p->x) << FP_SHIFT : p->x;
  tx = p2ht(roundedX);
  nx = (bool)nh(roundedX);  // true if entity overlaps right
  ty = p2vt(p->y);

  /* cell      = (GetTile(tx % VRAM_TILES_H, ty % VRAM_TILES_V) != TILE_GREEN); */
  /* cellright = (GetTile((tx + 1) % VRAM_TILES_H, ty % VRAM_TILES_V) != TILE_GREEN); */
  /* celldown  = (GetTile(tx % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V) != TILE_GREEN); */
  /* celldiag  = (GetTile((tx + 1) % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V) != TILE_GREEN); */

  cell      = isSolidForEntity(tx % VRAM_TILES_H, ty % VRAM_TILES_V, p->y, prevY, 0, p->down);
  cellright = isSolidForEntity((tx + 1) % VRAM_TILES_H, ty % VRAM_TILES_V, p->y, prevY, 0, p->down);
  celldown  = isSolidForEntity(tx % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V, p->y, prevY, 0, p->down);
  celldiag  = isSolidForEntity((tx + 1) % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V, p->y, prevY, 0, p->down);

  // The tests above comparing GetTile() to TILE_RED should be function calls
  // to isSolidForEntity() similar to what was done for Bugz (shown below)
  // This is how one-way tiles are implemented.

  /* offset = ty * SCREEN_TILES_H + tx; */
  /* cell      = isSolidForEntity(offset,                      ty,     prevY, WORLD_METER, p->down); // equiv. ... tx,     ty */
  /* cellright = isSolidForEntity(offset + 1,                  ty,     prevY, WORLD_METER, p->down); // equiv. ... tx + 1, ty */
  /* celldown  = isSolidForEntity(offset + SCREEN_TILES_H,     ty + 1, prevY, WORLD_METER, p->down); // equiv. ... tx,     ty + 1 */
  /* celldiag  = isSolidForEntity(offset + SCREEN_TILES_H + 1, ty + 1, prevY, WORLD_METER, p->down); // equiv. ... tx + 1, ty + 1 */

  if (p->dy > 0) {
    if ((      celldown && !cell) ||
        (nx && celldiag && !cellright)) {
      p->y = vt2p(p2vt(p->y));     // clamp the y position to avoid falling into platform below
      p->dy = 0;           // stop downward velocity
      p->jumping = false;  // no longer jumping
      p->framesFalling = 0;
    }
  } else if (p->dy < 0) {
    if ((      cell      && !celldown) ||
        (nx && cellright && !celldiag)) {
      p->y = vt2p(p2vt(p->y) + 1); // clamp the y position to avoid jumping into platform above
      p->dy = 0;           // stop upward velocity
    }
  }

  p->falling = !(celldown || (nx && celldiag)) && !p->jumping; // detect if we're now falling or not
  if (p->falling && p->framesFalling <= WORLD_FALLING_GRACE_FRAMES)
    p->framesFalling++;

  // Not needed for simple example
#if 0
  // Collision detection with ladders
  if (p->up || p->down) {
    if (p->down) {
      ty = p2vt(p->y + 1);     // the tile one sub-sub-pixel below current position
      ny = (bool)nv(p->y + 1); // true if player overlaps below
    } else { // p->up
      ty = p2vt(p->y - 1);     // the tile one sub-sub-pixel above current position
      ny = (bool)nv(p->y - 1); // true if entity overlaps above
    }

    uint16_t offset = ty * SCREEN_TILES_H + tx;
    if ((            isLadder(vram[offset                     ] - RAM_TILES_COUNT)) || // cell,      equiv. ... GetTile(tx,     ty    ) ...
        (nx &&       isLadder(vram[offset + 1                 ] - RAM_TILES_COUNT)) || // cellright, equiv. ... GetTile(tx + 1, ty    ) ...
        (ny &&       isLadder(vram[offset + SCREEN_TILES_H    ] - RAM_TILES_COUNT)) || // celldown,  equiv. ... GetTile(tx,     ty + 1) ...
        (nx && ny && isLadder(vram[offset + SCREEN_TILES_H + 1] - RAM_TILES_COUNT))) { // celldiag,  equiv. ... GetTile(tx + 1, ty + 1) ...
      if (p->down)
        p->y++; // allow entity to join a ladder directly below them
      else // p->up
        p->y--; // allow entity to join a ladder directly above them
      p->update = entity_update_ladder;
      p->animationFrameCounter = 0;
      p->jumping = false;
      p->dx = p->dy = 0;
    }
  }
#endif
}

void Player_render(PLAYER* const p, CAMERA* const c)
{
  static const char *map = map_walk0;
  static uint8_t flags = 0;

  if (p->attacking) {
    if ((p->animationFrameCounter % PLAYER_ATTACK_ANIMATION_FRAME_SKIP) == 0) {
      map = pgm_read_ptr_near(&attack_maps[pgm_read_byte(&playerAttackAnimation[p->animationFrameCounter / PLAYER_ATTACK_ANIMATION_FRAME_SKIP])]);
    }
    // Compile-time assert that we are working with a power of 2
    BUILD_BUG_ON(isNotPowerOf2(PLAYER_ATTACK_ANIMATION_FRAME_SKIP * NELEMS(playerAttackAnimation)));
    p->animationFrameCounter = (p->animationFrameCounter + 1) & (PLAYER_ATTACK_ANIMATION_FRAME_SKIP * NELEMS(playerAttackAnimation) - 1);
    goto skipped;
  }

  if (p->jumping || p->falling /*|| e->update == entity_update_flying*/) {
    if (p->dy >= 0)
      map = map_fall;
    else
      map = map_jump;
  } else {
    if (!p->left && !p->right && p->dx == 0) {
      map = map_walk0;
    } else {
      if ((p->animationFrameCounter % PLAYER_ANIMATION_FRAME_SKIP) == 0) {
	if (p->turbo && (p->dx == (WORLD_MAXDX + WORLD_METER) || p->dx == -(WORLD_MAXDX + WORLD_METER)))
	  map = pgm_read_ptr_near(&run_maps[pgm_read_byte(&playerAnimation[p->animationFrameCounter / PLAYER_ANIMATION_FRAME_SKIP])]);
	else
	  map = pgm_read_ptr_near(&walk_maps[pgm_read_byte(&playerAnimation[p->animationFrameCounter / PLAYER_ANIMATION_FRAME_SKIP])]);
      }
      // Compile-time assert that we are working with a power of 2
      BUILD_BUG_ON(isNotPowerOf2(PLAYER_ANIMATION_FRAME_SKIP * NELEMS(playerAnimation)));
      p->animationFrameCounter = (p->animationFrameCounter + 1) & (PLAYER_ANIMATION_FRAME_SKIP * NELEMS(playerAnimation) - 1);
    }
  }

 skipped:
  if (p->left)
    flags = SPRITE_FLIP_X;
  else if (p->right)
    flags = 0;

  // Round x and y to the nearest whole pixel for rendering purposes only
  MapSprite2(0, map, flags | SPRITE_BANK0);
  MoveSprite(0, nearestScreenPixel(p->x) - c->x - 8, nearestScreenPixel(p->y) - c->y - 24, 3, 4);

  /* MapSprite2(12, map, flags | SPRITE_BANK0); */
  /* MoveSprite(12, nearestScreenPixel(p->x) - c->x - 8 + 24, nearestScreenPixel(p->y) - c->y - 24, 3, 4); */
}

static void DrawHiddenTiles(CAMERA *c)
{
  int16_t cxt = c->x / TILE_WIDTH;
  int16_t cyt = c->y / TILE_HEIGHT;

  if (c->needColumn || c->needRow)
      SD_CompleteCued_Sector_Nr_Nc(&c->level->sds);

  // Have we scrolled past a tile boundary along X?
  if (c->needColumn) {
    if (c->forwardX) {
      // I think I might need to look at the rows/cols based on the previous camera position!
      uint8_t tx = (cxt + 29) % VRAM_TILES_H;
      for (uint8_t i = 0; i < VRAM_TILES_V; ++i) {
	SetTile(tx, ((uint16_t)cyt + i) % VRAM_TILES_V, c->level->buf[i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawColumn(c->level, cxt + 29, cyt, cxt + 29);
    } else {
      uint8_t tx = (cxt + 30) % VRAM_TILES_H;
      for (uint8_t i = 0; i < VRAM_TILES_V; ++i) {
	SetTile(tx, ((uint16_t)cyt + i) % VRAM_TILES_V, c->level->buf[32 + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + 32 + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawColumn(c->level, cxt + 30, cyt, cxt - 2);
    }
  }
  // Have we scrolled past a tile boundary along Y?
  if (c->needRow) {
    if (c->forwardY) {
      uint8_t ty = (cyt + 29) % VRAM_TILES_V;
      for (uint8_t i = 0; i < VRAM_TILES_H; ++i) {
	SetTile(((uint16_t)cxt + i) % VRAM_TILES_H, ty, c->level->buf[64 + i]);
	//SetTile(i, ty, c->buf[byteOffset + 64 + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + 64 + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawRow(c->level, cxt, cyt + 29, cyt + 29);
    } else {
      uint8_t ty = (cyt + 30) % VRAM_TILES_V;
      for (uint8_t i = 0; i < VRAM_TILES_H; ++i) {
	SetTile(((uint16_t)cxt + i) % VRAM_TILES_H, ty, c->level->buf[96 + i]);
	//SetTile(i, ty, c->buf[byteOffset + 64 + i]);
	/* UZEMHEX = c->sds.bufp[byteOffset + 64 + i]; */
	/* UZEMCHR = ' '; */
      }
      /* UZEMCHR = '\n'; */
      //Level_drawRow(c->level, cxt, cyt + 30, cyt - 2);
    }
  }

}

int main()
{
  SetTileTable(tileset);
  SetSpritesTileBank(0, mysprites);
  InitMusicPlayer(patches);

  CAMERA c;
  PLAYER p;
  LEVEL l;
  ClearVram();

  Level_initFromSD(&l);

  // SD card stuff
  /* sdc_struct_t sds; */
  /* uint8_t buf[512]; */
  /* uint32_t cluster; */

  /* c.sds.bufp = &c.buf[0]; */
  /* uint8_t res = FS_Init(&c.sds); */
  /* if (res != 0U) */
  /*   while (1); */

  /* cluster = FS_Find(&c.sds, */
  /* 	    ((u16)('S') << 8) | */
  /* 	    ((u16)('D')     ), */
  /* 	    ((u16)('L') << 8) | */
  /* 	    ((u16)('E')     ), */
  /* 	    ((u16)('V') << 8) | */
  /* 	    ((u16)('E')     ), */
  /* 	    ((u16)('L') << 8) | */
  /* 	    ((u16)('1')     ), */
  /* 	    ((u16)('D') << 8) | */
  /* 	    ((u16)('A')     ), */
  /* 	    ((u16)('T') << 8) | */
  /* 	    ((u16)(0)       )); */
  /* if (cluster == 0U) */
  /*   while (1); */
  /* FS_Select_Cluster(&c.sds, cluster); */
  /* c.startSector = FS_Get_Sector(&c.sds); */
  /* SPI_Set_Max(); */
  /* SDC_CRC_Enable(&c.sds, 0); */
  /*
   * To compare the output of this program with sdlevel1.dat:
   *
   *   ~/uzebox/bin/uzem sdmap.uze > log.txt
   *   hexdump -Cv sdlevel1.dat | cut -c 11-59 > hex.txt
   *   diff -uw log.txt hex.txt | less
   *
   */
  /* for (uint16_t i = 0; i < (60 - 28 + 1) * (56 - 28 + 1); ++i) { */
  /*   //SDC_Read_Sector_Fast_No_CRC(&c.sds, SDC_Command_Address(&c.sds, c.startSector + i)); */
  /*   SD_Read_Sector_Nr_Nc(&c.sds, FS_Get_Sector(&c.sds) + i); */
  /*   //FS_Read_Sector(&c.sds); */
  /*   for (uint16_t j = 0; j < 512;) { */
  /*     UZEMHEX = c.buf[j]; */
  /*     UZEMCHR = ' '; */
  /*     if ((++j % 8) == 0) */
  /* 	UZEMCHR = ' '; */
  /*     if ((j % 16) == 0) */
  /* 	UZEMCHR = '\n'; */
  /*   } */
  /* } */

  /* SetTile(0, 0, 1); */
  /* for (;;); */

btn_level:
  Camera_init(&c, &l);
  //Camera_fillVram(&c);
  Camera_fillVramFromSD(&c);
  Player_init(&p, &l);

  // Randomly place a bunch of collectibles
  /* for (uint8_t i = 0; i < 16; ++i) */
  /*   SetTile(random() % VRAM_TILES_H, random() % VRAM_TILES_V, TILE_GREEN); */
  __asm__ __volatile__ ("wdr");

  for (;;) {
    //__asm__ __volatile__ ("wdr");
    ProcessSprites();
    //__asm__ __volatile__ ("wdr");
    WaitVsync(1);
    RestoreBackground();
    ReadControllers();

    // This is where we can finish writing to VRAM if needed
    //__asm__ __volatile__ ("wdr");
    DrawHiddenTiles(&c);
    //__asm__ __volatile__ ("wdr");

    Player_input(&p);
    Player_update(&p);



  // This camera can follow the player, but it's too slow
  /* if (c.x < (p.x >> FP_SHIFT) - ((SCREEN_TILES_H*8)/2)+8) */
  /*   c.x+=1; */
  /* else if (c.x > (p.x >> FP_SHIFT) - ((SCREEN_TILES_H*8)/2)+8) */
  /*   c.x-=1; */

  /* if (!p.jumping && !p.falling) { */
  /*   if (c.y < (p.y >> FP_SHIFT) - ((SCREEN_TILES_V*8)/2)+8) */
  /*     c.y+=1; */
  /*   else if (c.y > (p.y >> FP_SHIFT) - ((SCREEN_TILES_V*8)/2)+8) */
  /*     c.y-=1; */
  /* } */

  /* Camera_moveTo(c, c.x, c.y); */

    int16_t targetX = nearestScreenPixel(p.x) - ((SCREEN_TILES_H * TILE_WIDTH) / 2) + TILE_WIDTH;
    int16_t targetY = nearestScreenPixel(p.y) - ((SCREEN_TILES_V * TILE_HEIGHT) / 2) + TILE_HEIGHT;
    Camera_moveTo(&c, targetX, targetY);

  // Combine the stuff above in various ways for a hybrid camera system
  /* Camera_moveTo(c, targetX, c.y); */


    //__asm__ __volatile__ ("wdr");
    Camera_update(&c);
    //__asm__ __volatile__ ("wdr");
    Player_render(&p, &c);


    // Figure out which tiles the player sprite overlaps with
    /* uint8_t tx = p2ht(p.x); */
    /* bool nx = (bool)nh(p.x); // true if entity overlaps right */
    /* uint8_t ty = p2vt(p.y); */
    /* bool ny = (bool)nv(p.y); // true if entity overlaps below */

    /* if (GetTile(tx % VRAM_TILES_H, ty % VRAM_TILES_V) == TILE_GREEN) { */
    /*   TriggerFx(1, 128, true); */
    /*   SetTile(tx % VRAM_TILES_H, ty % VRAM_TILES_V, TILE_BLACK); */
    /* } */
    /* if (nx) { */
    /*   if (GetTile((tx + 1) % VRAM_TILES_H, ty % VRAM_TILES_V) == TILE_GREEN) { */
    /*     TriggerFx(1, 128, true); */
    /*     SetTile((tx + 1) % VRAM_TILES_H, ty % VRAM_TILES_V, TILE_BLACK); */
    /*   } */
    /* } */
    /* if (ny) { */
    /*   if (GetTile(tx % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V) == TILE_GREEN) { */
    /*     TriggerFx(1, 128, true); */
    /*     SetTile(tx % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V, TILE_BLACK); */
    /*   } */
    /* } */
    /* if (nx && ny) { */
    /*   if (GetTile((tx + 1) % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V) == TILE_GREEN) { */
    /*     TriggerFx(1, 128, true); */
    /*     SetTile((tx + 1) % VRAM_TILES_H, (ty + 1) % VRAM_TILES_V, TILE_BLACK); */
    /*   } */
    /* } */

    if (p.buttons.released & BTN_START) {
      //Level_initFromMap(&l, map_level);
      Level_initFromSD(&l);
      goto btn_level;
    }

    if (p.buttons.released & BTN_SELECT) {
      UZEMCHR = 'X'; UZEMCHR = ' '; UZEMHEX = HI8((uint16_t)p2ht(p.x)); UZEMHEX = LO8((uint16_t)p2ht(p.x)); UZEMCHR = ','; UZEMCHR = ' ';
      UZEMCHR = 'Y'; UZEMCHR = ' '; UZEMHEX = HI8((uint16_t)p2vt(p.y)); UZEMHEX = LO8((uint16_t)p2vt(p.y)); UZEMCHR = '\n';
    }

    /* if ((p.buttons.released & BTN_SR)) { */
    /*   uint8_t level = l.level + 1; */
    /*   if (l.level == NUM_LEVELS - 1) */
    /* 	level = 0; */
    /*   Level_init(&l, level); */
    /*   goto btn_level; */
    /* } */

    /* if ((p.buttons.released & BTN_SL)) { */
    /*   uint8_t level = l.level - 1; */
    /*   if (l.level == 0) */
    /* 	level = NUM_LEVELS - 1; */
    /*   Level_init(&l, level); */
    /*   goto btn_level; */
    /* } */
  }
}
